package ppi.app;

import ppi.bsets.RelationalBayesianSetsBCMerged;
import ppi.bsets.BayesianSetsGaussian;

import java.io.IOException;
import java.io.BufferedReader;
import java.io.FileReader;
import java.util.*;

/**
 * HOW DOES THIS WORK?
 *
 * This is what one needs to begin with:
 *
 * - a PROTEIN_ID_FILE: the role of this file is just to list the protein IDs, which are basically strings
 *   with the protein name;
 * - a GRAPH_DATA_FILE: specifies which proteins interact. The numbers on this file correspond to the protein
 *   line number in the PROTEIN_ID_FILE, starting from 0 (this is stored in the proteinIndex hashtable).
 *   This graph usually contains the same information in both  ways, i.e., interaction A:B is also shown as
 *   interaction B:A;
 * - a PROTEIN_PROTEINCAT_FILE: a categorization of the function of the proteins. Usually, taken out of the
 *   gene ontology categotization;
 * - a PRIORS_NAME file: namely, the mean parameters of a multivariate Gaussian
 *   prior. Each line contains a different parameter.
 * - a FEATURES_DATA_FILE: contains the predictive features for each pair of protein (one pair per line).
 *   These features are symmetric with respect to the pair.
 * - a EXTERNAL_POSITIVE_PAIRS file: indicates the line number in the FEATURES_DATA_FILE for each pair in this file.
 *   For instance, "YAL021C	YCR093W	1" is the 11th line in the file, indicating that the 11th line in
 *   FEATURES_DATA_FILE contains the respective data for this file.
 *
 * How to customize this program:
 *
 *  There are two indices of proteins pairs, which can sound a bit confusing. The reason for that is that we might
 *  be inquirying about links in a particular graph (links[][]) but drawing data from a table (FEATURES_DATA_FILE)
 *  that is indexed by another graph (externalTablePositive). The common index system between both data sources is
 *  controlled by the string ids of the proteins (within proteinNames and externalTablePositive). Notice we only
 *  load the positive links in externalTablePositive.
 *
 *  This separation allows us to load data from a particular data source (indexed by EXTERNAL_POSITIVE_PAIRS) but do queries
 *  with a separate selection of proteins (GRAPH_DATA_FILE/PROTEIN_ID_FILE).
 *
 */

public class PPIQuery {

    private static String  PROTEIN_ID_FILE = "mips.protein";
    private static String  PROTEIN_PROTEINCAT_FILE = "mipsCat.txt";
    private static String  GRAPH_DATA_FILE = "mips.graph";
    private static String  EXTERNAL_POSITIVE_PAIRS = "mipsPosPair";

    private static String  QUERY_FILE = null, LABEL_1 = null, LABEL_2 = null;
    private static int     DEFAULT_QUERY_SIZE = 15;
    private static double  DEFAULT_C = 1.;

    private static String  PRIORS_NAME = "mips.priors";
    private static String  FEATURES_DATA_FILE = "mipsPosPair.feature";

    private static boolean VERBOSE = false;
    private static boolean FIRST_HUNDRED = false;

    public static void main(final String[] argv) {

        System.out.println("** Protein-protein interaction retrieval");
        System.out.println(" - Code by Ricardo Silva, Gatsby Computational Neuroscience Unit, 2007");
        System.out.println(" - Last modification: Feb 02 2007");
        System.out.println(" - Usage: java ppi.app -querytype <mips type 1> <mips type 2> [-queryfile <file name>]" +
                              " [-firsthundred");
        System.out.println();

        int argc = 0;
        while (argc < argv.length) {
            if (argv[argc].equals("-queryfile")) {
                if (argc < argv.length - 1) {
                    argc++;
                    QUERY_FILE = argv[argc];
                } else {
                    System.out.println("Missing argument for -queryfile!");
                    System.exit(0);
                }
            } else if (argv[argc].equals("-querytype")) {
                if (argc < argv.length - 2) {
                    LABEL_1 = argv[++argc];
                    LABEL_2 = argv[++argc];
                    while (LABEL_1.startsWith("0"))
                        LABEL_1 = LABEL_1.substring(1);
                    while (LABEL_2.startsWith("0"))
                        LABEL_2 = LABEL_2.substring(1);
                } else {
                    System.out.println("Missing argument for -querytype!");
                    System.exit(0);
                }
            } else if (argv[argc].equals("-firsthundred")) {
                FIRST_HUNDRED = true;
            } else if (argv[argc].equals("-c")) {
                if (argc < argv.length - 1) {
                    argc++;
                    DEFAULT_C = Double.parseDouble(argv[argc]);
                } else {
                    System.out.println("Missing argument for -c!");
                    System.exit(0);
                }
            }else {
                System.out.println("Invalid argument: " + argv[argc]);
                System.exit(0);
            }
            argc++;
        }

        if (LABEL_1 == null || LABEL_2 == null) {
            System.out.println("Missing query type information! Use -querytype");
            System.exit(0);
        }

        try {
            //Load the main indices: the graph which contains the positive class information (links[][]) with
            //the respective protein ids (proteinNames[])
            String proteinNames[] = readMainProteins(PROTEIN_ID_FILE);
            int    links[][] = loadMainLinksCompact(GRAPH_DATA_FILE, proteinNames.length);
            Hashtable<String, Integer> proteinIndex = new Hashtable<String, Integer>();
            for (int i = 0; i < proteinNames.length; i++)
                 proteinIndex.put(proteinNames[i], i);
            Hashtable<String, Integer> proteinIds = new Hashtable<String, Integer>();
            for (int i = 0; i < proteinNames.length; i++)
                 proteinIds.put(proteinNames[i], i);
            List<String>[] proteinClasses = getProteinClasses(proteinIndex, proteinNames);
            Hashtable<String, List<String>> classDistribution = getClassDistribution(proteinNames, proteinClasses);

            //Load the feature data (data[][]) with the index system (pairIndex) that allows us to find protein pair
            //data selected within proteinNames/links
            double data[][] = loadMipsFeatures(FEATURES_DATA_FILE);
            List externalTablePositive = loadExternalLinks(EXTERNAL_POSITIVE_PAIRS);
            Hashtable<Integer, Hashtable<Integer, Integer>> pairIndex = getPairIndex(proteinIds,
                        (Hashtable<String, Hashtable<String, Integer>>) externalTablePositive.get(0));

            //Do the actual retrieval
            retrievalTest(data, links, proteinClasses, proteinNames, classDistribution, proteinIds, pairIndex);
            System.out.println("Done!");
        } catch (IOException e) {
            System.out.println("File not loaded! " + e.getMessage());
        }
    }

    public static String[] readMainProteins(String fileName) throws java.io.IOException {
        BufferedReader in = new BufferedReader(new FileReader(fileName));
        List<String> preIds = new ArrayList<String>();
        while (true) {
            String line = in.readLine();
            if (line == null)
                break;
            preIds.add(line.trim());
        }
        in.close();
        String output[] = new String[preIds.size()];
        for (int i = 0; i < preIds.size(); i++)
            output[i] = preIds.get(i);
        return output;
    }

    protected static Hashtable<Integer, Hashtable<Integer, Integer>> getPairIndex(Hashtable<String, Integer> proteinIds,
    Hashtable<String, Hashtable<String, Integer>> stringIndex) {
        Hashtable<Integer, Hashtable<Integer, Integer>> output = new Hashtable<Integer, Hashtable<Integer, Integer>>();
        for (String name1 : stringIndex.keySet()) {
            Integer position1 = proteinIds.get(name1);
            Hashtable<Integer, Integer> tableName1 = new Hashtable<Integer, Integer>();
            Hashtable<String, Integer> substringIndex = stringIndex.get(name1); 
            for (String name2 : substringIndex.keySet()) {
                Integer position2 = proteinIds.get(name2);
                tableName1.put(position2, substringIndex.get(name2));
            }
            output.put(position1, tableName1);
        }
        return output;
   }

    protected static int[][] loadMainLinksCompact(String fileName, int size) throws java.io.IOException {
        int links[][] = new int[size][], numLinks = 0;
        List<Integer> setLinks[] = new ArrayList[size];
        for (int i = 0; i < size; i++)
            setLinks[i] = new ArrayList<Integer>();
        BufferedReader file = new BufferedReader(new FileReader(fileName));
        String line = file.readLine();
        String doc[] = line.split(" ");
        for (int i = 1; i < doc.length; i++) {
            String sub[] = doc[i].split(":");
            int p1 = Integer.valueOf(sub[0]);
            int p2 = Integer.valueOf(sub[1]);
            if (p1 > p2)
                continue;
            setLinks[p1].add(p2);
            numLinks++;
        }
        System.out.println("TOTAL NUMBER OF LINKS = " + numLinks);
        for (int i = 0; i < size; i++) {
            links[i] = new int[setLinks[i].size()];
            int pos = 0;
            for (Integer key : setLinks[i])
                links[i][pos++] = key;
        }
        return links;
    }

    protected static double[][] loadMipsFeatures(String fileName) throws java.io.IOException {
        List<double[]> rawData = new ArrayList<double[]>();
        BufferedReader file = new BufferedReader(new FileReader(fileName));
        String line = file.readLine();
        int lineNumber = 1;
        while (line != null) {
            String array[] = line.split(",");
            if (lineNumber % 1000 == 0)
                System.out.println("Reading line #" + lineNumber);
            lineNumber++;
            double newPoint[] = new double[41];
            double norm = 0., obsMean = 0.;
            int    numRecorded = 0;
            for (int i = 0; i < 20; i++) {
                newPoint[i] = Double.parseDouble(array[i]);
                if (newPoint[i] != -100) {
                    numRecorded++;
                    obsMean += newPoint[i];
                }
            }
            obsMean /= numRecorded;
            for (int i = 0; i < 20; i++) {
                if (newPoint[i] == -100)
                    newPoint[i] = obsMean;
                norm += newPoint[i] * newPoint[i];
            }
            norm = Math.sqrt(norm); //for non-linear transformation
            for (int i = 0; i < 20; i++) {
                newPoint[20 + i] = newPoint[i] * newPoint[i] / norm;
                newPoint[i] /= norm;
            }
            newPoint[40] = 1.;
            rawData.add(newPoint);
            line = file.readLine();
        }
        double dataArray[][] = new double[rawData.size()][];
        for (int i = 0; i < dataArray.length; i++)
             dataArray[i] = rawData.get(i);
        return dataArray;
    }

    protected static List<String>[] getProteinClasses(Hashtable<String, Integer> proteinIds, String proteinNames[])
    throws java.io.IOException {
        String fileName = PROTEIN_PROTEINCAT_FILE;
        List<String> proteinClasses[] = new ArrayList[proteinIds.size()];
        for (int i = 0; i < proteinClasses.length; i++)
            proteinClasses[i] = new ArrayList<String>();
        BufferedReader file = new BufferedReader(new FileReader(fileName));
        file.readLine();
        while (true) {
            String line = file.readLine(); //get category name
            if (line == null)
                break;
            file.readLine();
            String name = processedClassName(line);
            //Now read protein names
            line = file.readLine();
            while (line != null && !line.equals("")) {
                if (proteinIds.get(line) == null) {
                    line = file.readLine();
                    continue;
                }
                int code = proteinIds.get(line);
                proteinClasses[code].add(name);
                line = file.readLine();
            }
            file.readLine();
        }
        file.close();
        System.out.println("Summary:");
        for (int i = 0; i < proteinClasses.length; i++) {
            System.out.println("PROTEIN " + proteinNames[i]);
            System.out.println(" - " + proteinClasses[i].size() + " classes");
            System.out.println(" - " + getClassesString(i, proteinClasses));
            proteinClasses[i] = filterClasses(proteinClasses[i]);
            System.out.println(" - " + getClassesString(i, proteinClasses));
            System.out.println();
        }
        return proteinClasses;
    }

    private static String processedClassName(String input) {
        String tokens[] = input.split(" +()\"");
        StringBuffer name = new StringBuffer();
        name.append(tokens[0].toLowerCase());
        for (int i = 1; i < tokens.length; i++) {
            if (tokens[i].equals("(") && (tokens[i + 2].equals("ORF") || tokens[i + 2].equals("ORFs")))
                break;
            name.append(" " + tokens[i].toLowerCase());
        }
        return new String(name);
    }

    protected static List<String> filterClasses(List<String> original) {
        String[][] tokenizedOutput = new String[original.size()][];
        int count = 0;
        for (Iterator<String> it  = original.iterator(); it.hasNext();) {
            String item = it.next();
            String realItem;
            if (item.charAt(0) == '0')
                realItem = item.substring(1);
            else
                realItem = item;
            String tokens[] = realItem.split("\\D");
            tokenizedOutput[count++] = tokens;
        }
        List<String> output = new ArrayList<String>();
        nextClass:
        for (int i = 0; i < tokenizedOutput.length; i++) {
            for (int j = i + 1; j < tokenizedOutput.length; j++)
                if (tokenizedOutput[i].length < tokenizedOutput[j].length) {
                    boolean match = true;
                    for (int k = 0; k < tokenizedOutput[i].length; k++)
                        if (!tokenizedOutput[i][k].equals(tokenizedOutput[j][k])) {
                            match = false;
                            break;
                        }
                    if (match)
                        continue nextClass;
                }
            StringBuffer approved = new StringBuffer();
            for (int j = 0; j < tokenizedOutput[i].length - 1; j++)
                approved.append(tokenizedOutput[i][j] + ".");
            approved.append(tokenizedOutput[i][tokenizedOutput[i].length - 1]);
            output.add(new String(approved));
        }
        return output;
    }

    protected static List loadExternalLinks(String fileName) throws java.io.IOException {
        Hashtable<String, Hashtable<String, Integer>> externalOutput1 = new Hashtable<String, Hashtable<String, Integer>>();
        Hashtable<String, HashSet<String>> externalOutput2 = new Hashtable<String, HashSet<String>>();
        BufferedReader file = new BufferedReader(new FileReader(fileName));
        int numLinks = 0;
        while (true) {
            String line = file.readLine();
            if (line == null)
                break;
            String doc[] = line.split("\t");
            for (int d = 0; d < 2; d++) {
                if (doc[d].contains("-")) {
                    int pos = doc[d].indexOf('-');
                    doc[d] = doc[d].substring(0, pos) + doc[d].substring(pos + 1, doc[d].length());
                }
            }

            Hashtable<String, Integer> values1 = externalOutput1.get(doc[0]);
            if (values1 == null) {
                values1 = new Hashtable<String, Integer>();
                externalOutput1.put(doc[0], values1);
            }
            values1.put(doc[1], numLinks);
            values1 = externalOutput1.get(doc[1]);
            if (values1 == null) {
                values1 = new Hashtable<String, Integer>();
                externalOutput1.put(doc[1], values1);
            }
            values1.put(doc[0], numLinks);

            HashSet<String> values2 = externalOutput2.get(doc[0]);
            if (values2 == null) {
                values2 = new HashSet<String>();
                externalOutput2.put(doc[0], values2);
            }
            values2.add(doc[1]);
            values2 = externalOutput2.get(doc[1]);
            if (values2 == null) {
                values2 = new HashSet<String>();
                externalOutput2.put(doc[1], values2);
            }
            values2.add(doc[0]);

            numLinks++;
            if (numLinks % 1000 == 0)
                System.out.println("#links so far = " + numLinks);
        }
        System.out.println("TOTAL NUMBER OF LINKS = " + numLinks);
        List output = new ArrayList(); output.add(externalOutput1); output.add(externalOutput2);
        return output;
    }

    protected static int[][] getPartiallyRandomQuery(int links[][], String class1, String class2,
                                                     Hashtable<String, List<String>> classDistribution,
                                                     Hashtable<String, Integer> proteinIds,
                                                     String proteinNames[], int querySize, double dataArray[][],
                                                     Hashtable<Integer, Hashtable<Integer, Integer>> pairsIndex,
                                                     boolean allowedRow[]) {
        List<String> proteins1 = classDistribution.get(class1);
        List<String> proteins2 = classDistribution.get(class2);
        List<String> allPairs1[] = new ArrayList[proteins1.size()];
        List<String> allPairs2[] = new ArrayList[proteins2.size()];
        boolean selected1[][] = new boolean[proteins1.size()][];
        boolean selected2[][] = new boolean[proteins2.size()][];
        for (int i = 0; i < allPairs1.length; i++) {
            List<String> related = new ArrayList<String>();
            int idx = proteinIds.get(proteins1.get(i));
            for (int j = 0; j < links[idx].length; j++)
                if (proteins2.contains(proteinNames[links[idx][j]]))
                    related.add(proteinNames[links[idx][j]]);
            allPairs1[i] = related;
            selected1[i] = new boolean[related.size()];
            for (int j = 0; j < selected1[i].length; j++)
                selected1[i][j] = false;
        }
        for (int i = 0; i < allPairs2.length; i++) {
            List<String> related = new ArrayList<String>();
            int idx = proteinIds.get(proteins2.get(i));
            for (int j = 0; j < links[idx].length; j++)
                if (proteins1.contains(proteinNames[links[idx][j]]))
                    related.add(proteinNames[links[idx][j]]);
            allPairs2[i] = related;
            selected2[i] = new boolean[related.size()];
            for (int j = 0; j < selected2[i].length; j++)
                selected2[i][j] = false;
        }
        int count = 0;
        int output[][] = new int[querySize][2];
        Random r = new Random();
        mainSelection:
        while (count < querySize) {
            boolean selected[][];
            List<String> p1, allPairs[];
            if (r.nextDouble() < ((double) selected1.length) / (selected1.length + selected2.length)) {
                selected = selected1;
                p1 = proteins1;
                allPairs = allPairs1;
            }
            else {
                selected = selected2;
                p1 = proteins2;
                allPairs = allPairs2;
            }
            int row = r.nextInt(p1.size());
            if (selected[row].length == 0)
                continue;
            int column = r.nextInt(selected[row].length);
            for (int i = 0; i < count; i++) {
                if (output[i][0] == proteinIds.get(p1.get(row)) && output[i][1] == proteinIds.get(allPairs[row].get(column))) {
                    System.out.println("Repeated: try again!");
                    continue mainSelection;
                }
                if (output[i][0] == proteinIds.get(allPairs[row].get(column)) && output[i][1] == proteinIds.get(p1.get(row))) {
                    System.out.println("Repeated: try again!");
                    continue mainSelection;
                }
            }
            double point[] = buildDataPoint(proteinIds.get(p1.get(row)), proteinIds.get(allPairs[row].get(column)), dataArray, pairsIndex);
            if (point == null)
                continue;
            if (!allowedRow[pairsIndex.get(proteinIds.get(p1.get(row))).get(proteinIds.get(allPairs[row].get(column)))]) {
                System.out.println("* NOT ALLOWED *: try again!");
                continue mainSelection;
            }
            output[count][0] = proteinIds.get(p1.get(row));
            output[count][1] = proteinIds.get(allPairs[row].get(column));
            System.out.println("Picking (" + p1.get(row) + " x " + allPairs[row].get(column) + ")");
            int idx = pairsIndex.get(output[count][0]).get(output[count][1]);
            for (int i = 0; i < 10; i++)
                System.out.print(dataArray[idx][i] + " ");
            System.out.println();
            selected[row][column] = true;
            count++;
        }
        return output;
    }

    protected static double[] buildDataPoint(int d1, int d2, double data[][], Hashtable<Integer, Hashtable<Integer, Integer>> pairsIndex) {
        Hashtable<Integer, Integer> idx1 = pairsIndex.get(d1);
        if (idx1 == null)
            return null;
        Integer idx2 = pairsIndex.get(d1).get(d2);
        if (idx2 == null)
            return null;
        if (Double.isNaN(data[idx2][0]))
            return null;
        return data[idx2];
    }
    
    protected static void printQuery(int query[][], List<String> proteinClasses[], String proteinNames[]) {
        for (int i = 0; i < query.length; i++) {
            System.out.println("Query " + i +":");
            for (int p = 0; p < 2; p++) {
                System.out.print(" - " + proteinNames[query[i][p]] + " (");
                for (int j = 0; j < proteinClasses[query[i][p]].size() - 1; j++)
                    System.out.print(proteinClasses[query[i][p]].get(j) + ", ");
                System.out.println(proteinClasses[query[i][p]].get(proteinClasses[query[i][p]].size() - 1) + ")");
            }
        }
    }

    protected static Hashtable<String, List<String>> getClassDistribution(String proteinNames[], List<String> proteinClasses[]) {
        Hashtable<String, List<String>> directory = new Hashtable<String, List<String>>();
        HashSet<String> classSet = new HashSet<String>();
        for (int i = 0; i < proteinClasses.length; i++) {
            for (Iterator<String> it = proteinClasses[i].iterator(); it.hasNext();) {
                String p = it.next();
                if (!classSet.contains(p))
                    classSet.add(p);
            }
        }
        for (Iterator<String> it = classSet.iterator(); it.hasNext();)
            directory.put(it.next(), new ArrayList<String>());
        for (int i = 0; i < proteinClasses.length; i++) {
            for (Iterator<String> it = proteinClasses[i].iterator(); it.hasNext();) {
                List<String> list = directory.get(it.next());
                list.add(proteinNames[i]);
            }
        }
        if (VERBOSE)
            for (Iterator<String> it = directory.keySet().iterator(); it.hasNext();) {
                String className = it.next();
                List<String> proteins = directory.get(className);
                System.out.println("CLASS " + className +  ", size = " + proteins.size());
                for (Iterator it2 = proteins.iterator(); it2.hasNext();)
                    System.out.println(" - " + it2.next());
            }
        return directory;
    }

    protected static String getClassesString(int id, List<String> proteinClasses[]) {
        StringBuffer output = new StringBuffer();
        for (Iterator it = proteinClasses[id].iterator(); it.hasNext();)
            output.append(it.next() + " ");
        return new String(output);
    }

    protected static boolean inQuery(int query[][], int v1, int v2) {
        for (int q = 0; q < query.length; q++)
            if ((query[q][0] == v1 && query[q][1] == v2) || (query[q][0] == v2 && query[q][1] == v1))
                return true;
        return false;
    }

    protected static int[][] readQueryFromFile(String fileName, Hashtable<String, Integer> proteinIds) {
        try {
            BufferedReader in = new BufferedReader(new FileReader(fileName));
            String line = in.readLine();
            List<int[]> pointList = new ArrayList<int[]>();
            String names[] = new String[2];
            while (line != null) {
                names[0] = line.trim().split("\t| ")[0];
                names[1] = line.trim().split("\t| ")[1];
                int point[] = {proteinIds.get(names[0]), proteinIds.get(names[1])};
                pointList.add(point);
                line = in.readLine();
            }
            int points[][] = new int[pointList.size()][];
            for (int i = 0; i < points.length; i++)
                points[i] = pointList.get(i);
            return points;
        } catch (IOException e) {
            System.out.println("File not loaded! " + e.getMessage());
        }
        return null;
    }

    /**
     * Parameters
     *
     * @param dataArray: features for each pair
     * @param links:
     * @param proteinClasses
     * @param proteinNames
     * @param proteinIds
     * @param pairsIndex
     */

    protected static void retrievalTest(double dataArray[][], int links[][], List<String> proteinClasses[],
                                        String proteinNames[], Hashtable<String, List<String>> classDistribution,
                                        Hashtable<String, Integer> proteinIds,
                                        Hashtable<Integer, Hashtable<Integer, Integer>> pairsIndex) {
        int notAllowed = 0;
        boolean allowedRow[] = new boolean[dataArray.length];
        nextRow:
        for (int i = 0; i < allowedRow.length; i++) {
            //Extra: do not allow for missing values
            boolean foundOne = false;
            for (int j = 0; j < dataArray[i].length; j++)
                if (dataArray[i][j] == -100.) {
                    allowedRow[i] = false;
                    notAllowed++;
                    continue nextRow;
                }
            for (int j = 0; j < 20; j++) {
                if (Double.isNaN(dataArray[i][j])) {
                    allowedRow[i] = false;
                    notAllowed++;
                    continue nextRow;
                }
                for (int k = j + 1; k < 20; k++)
                    if (dataArray[i][j] == dataArray[i][k]) {
                        if (!foundOne) {
                            foundOne = true; //Allows for up to 1 repetition
                            continue;
                        }
                        allowedRow[i] = false;
                        notAllowed++;
                        continue nextRow;
                    }
            }
            allowedRow[i] = true;
        }
        System.out.println("Number of forbidden rows: " + notAllowed);

        int query[][];
        if (QUERY_FILE == null)
            query = getPartiallyRandomQuery(links, LABEL_1, LABEL_2, classDistribution, proteinIds, proteinNames,
                                            DEFAULT_QUERY_SIZE, dataArray, pairsIndex, allowedRow);
        else
            query = readQueryFromFile(QUERY_FILE, proteinIds);
        printQuery(query, proteinClasses, proteinNames);

        boolean allowableTypeI[] = new boolean[proteinNames.length];
        for (int i = 0; i < proteinNames.length; i++) {
            allowableTypeI[i] = proteinClasses[i].size() != 0;
            if (proteinNames[i].equals("YDL184WB")) //For some reason, not on Qi's data
                allowableTypeI[i] = false;
        }
        boolean allowableTypeII[] = new boolean[dataArray.length];
        for (int i = 0; i < proteinNames.length; i++) {
            allowableTypeII[i] = proteinClasses[i].size() != 0;
            if (proteinNames[i].equals("YDL184WB")) //For some reason, not on Qi's data
                allowableTypeII[i] = false;
        }

        Date date1 = new Date();

        RelationalBayesianSetsBCMerged rbSets = new RelationalBayesianSetsBCMerged(proteinNames.length, dataArray,
                           links, DEFAULT_C, PRIORS_NAME, pairsIndex, proteinNames);
        int resultsRBSets[][] = rbSets.query(query, allowableTypeI, allowableTypeII, null, allowedRow);
        resultsReport("*> RELATIONAL BAYESIAN SETS ", resultsRBSets, proteinNames, LABEL_1, LABEL_2, proteinClasses,
                                          query, pairsIndex, allowableTypeI, allowableTypeII, allowedRow, dataArray);
        System.out.println(resultsRBSets.length);
        System.out.println();

        double gbsetsData[][] = new double[dataArray.length][20];
        for (int d = 0; d < dataArray.length; d++) {
            int dPointer = 0;
            for (int i = 0; i < 20; i++)
                 gbsetsData[d][dPointer++] = dataArray[d][i];
        }
        BayesianSetsGaussian gBSets = new BayesianSetsGaussian(gbsetsData, true, allowedRow);
        double queryData[][] = new double[query.length][];
        for (int i = 0; i < queryData.length; i++)
            queryData[i] = gbsetsData[(pairsIndex.get(query[i][0])).get(query[i][1])];
        double resultsGBSetsRow[][] = gBSets.query(queryData);
        int inverseIndex[][] = new int[dataArray.length][2];
        for (Integer i : pairsIndex.keySet()) {
            Hashtable<Integer, Integer> rowIdx = pairsIndex.get(i);
            for (Integer j : rowIdx.keySet()) {
                int row = rowIdx.get(j);
                inverseIndex[row][0] = i; inverseIndex[row][1] = j;
            }
        }
        int resultsGBSets[][] = new int[resultsGBSetsRow.length][];
        for (int i = 0; i < resultsGBSets.length; i++)
            resultsGBSets[i] = inverseIndex[(int) resultsGBSetsRow[i][0]];
        resultsReport("*> GAUSSIAN BAYESIAN SETS ", resultsGBSets, proteinNames, LABEL_1, LABEL_2, proteinClasses,
                      query, pairsIndex, allowableTypeI, allowableTypeII, allowedRow, dataArray);
        System.out.println(resultsGBSets.length);
        System.out.println();

        RelationalBayesianSetsBCMerged rbSetsCosine = new RelationalBayesianSetsBCMerged(proteinNames.length, dataArray,
                                                          links, DEFAULT_C, PRIORS_NAME, pairsIndex, proteinNames);
        int resultsCosine[][] = rbSetsCosine.queryByCosine(query, allowableTypeI, allowableTypeII, null, allowedRow);
        resultsReport("*> COSINE DISTANCE", resultsCosine, proteinNames, LABEL_1, LABEL_2, proteinClasses, query,
                      pairsIndex, allowableTypeI, allowableTypeII, allowedRow, dataArray);
        System.out.println();

        RelationalBayesianSetsBCMerged rbSetsNN = new RelationalBayesianSetsBCMerged(proteinNames.length, dataArray,
                                                      links, DEFAULT_C, PRIORS_NAME, pairsIndex, proteinNames);
        int resultsNN[][] = rbSetsNN.queryByNearestNeighborsPlain(query, allowableTypeI, allowableTypeII, null, allowedRow);
        resultsReport("*> NEAREST NEIGHBORS", resultsNN, proteinNames, LABEL_1, LABEL_2, proteinClasses, query,
                      pairsIndex, allowableTypeI, allowableTypeII, allowedRow, dataArray);
        System.out.println(resultsNN.length);
        System.out.println();

        System.out.println("Original query: " + LABEL_1 + ", " + LABEL_2 + " (size = " + query.length + ")");
        Date date2 = new Date();
        System.out.println("Started at " + date1.toString() + ",finished at " + date2.toString());
    }

   protected static void resultsReport(String title, int results[][], String proteinNames[],
                                        String label1, String label2, List<String> proteinClasses[], int query[][],
                                        Hashtable<Integer, Hashtable<Integer, Integer>> pairsIndex,
                                        boolean allowableI[], boolean allowableII[], boolean allowedRow[], double dataArray[][]) {
        double hitPoints[] = new double[results.length];
        int    countable[] = new int[results.length];
        for (int v = 0; v < results.length; v++) {
            hitPoints[v] = 0; countable[v] = 0;
            String classes1 = getClassesString(results[v][0], proteinClasses);
            String classes2 = getClassesString(results[v][1], proteinClasses);
            boolean in = inQuery(query, results[v][0], results[v][1]);
            if (results[v][0] == results[v][1])
                continue;
            int idx = pairsIndex.get(results[v][0]).get(results[v][1]);
            if (!allowedRow[idx])
                continue;
            if (!allowableI[results[v][0]] || !allowableII[results[v][1]])
                continue;
            if (v < 100) {
                System.out.println("----> Pair #" + (v + 1) + "[index = " + pairsIndex.get(results[v][0]).get(results[v][1]) + "]");
                System.out.println(" - " + proteinNames[results[v][0]] + ": " + classes1);
                System.out.println(" - " + proteinNames[results[v][1]] + ": " + classes2);
                if (v < 50) {
                    for (int i = 0; i < 10; i++)
                        System.out.print(dataArray[idx][i] + " ");
                    System.out.println();
                }
            }
            if (in) {
                System.out.println("IN QUERY! Position = " + v);
                continue;
            }
            if ((classes1.contains(label1) && classes2.contains(label2)) || (classes1.contains(label2) && classes2.contains(label1)))
                hitPoints[v] = 1;
            countable[v] = 1;
        }
        double maxPoints = 0.;
        for (int i = 0; i < results.length; i++)
            if (countable[i] > 0)
                maxPoints += hitPoints[i];
        System.out.println(); System.out.println("Maximum number of points: " + maxPoints);
        System.out.println("---- PRECISION / RECALL FOR " + title);
        double recalled = 0, counted = 0;
        for (int i = 0; i < results.length; i++) {
            if (countable[i] == 1) {
                recalled += hitPoints[i];
                if (!FIRST_HUNDRED || counted < 100)
                    System.out.println(counted + ": " + recalled / (counted + 1) + "\t" + recalled / maxPoints);
                counted++;
            }
        }
    }

}

