package ppi.bsets;

import cern.colt.matrix.linalg.Algebra;
import cern.colt.matrix.DoubleMatrix2D;
import cern.colt.matrix.impl.DenseDoubleMatrix2D;

import java.util.Arrays;

/**
 * Bayesian sets of multivariate Gaussian variables, as in Ghahramani and Heller (2005).
 *
 * This does NOT assume independence between variables. Representation is DENSE.
 *
 * It also uses my sparse matrix representation of the relational version.
 *
 */

public class BayesianSetsGaussian {

    int     numVars;
    double  data[][];
    double  priorMean[], priorInvCov[][], priorCov[][], alpha, upsilon;
    double  posteriorMean[], posteriorInvCov[][], posteriorAlpha, posteriorUpsilon;
    boolean validRow[];

    public BayesianSetsGaussian(double data[][], boolean fitH) {
        this.validRow = null;
        initialization(data, fitH);
    }

    public BayesianSetsGaussian(double data[][], boolean fitH, boolean validRow[]) {
        this.validRow = validRow;
        initialization(data, fitH);
    }

    protected void initialization(double data[][], boolean fitH) {
        this.numVars = data[0].length;
        this.data = data;
        this.priorMean = new double[this.numVars];
        this.priorInvCov= new double[this.numVars][this.numVars];
        this.priorCov= new double[this.numVars][this.numVars];
        this.posteriorMean = new double[this.numVars];
        this.posteriorInvCov= new double[this.numVars][this.numVars];
        if (fitH)
            fitHyperparameters();
    }

    protected void fitHyperparameters() {
        for (int i = 0; i < this.numVars; i++) {
            this.priorMean[i] = 0.;
            for (int j = 0; j < this.numVars; j++)
                this.priorCov[i][j] = 0.;
        }
        int numValid = 0;
        for (int d = 0; d < this.data.length; d++) {
            if (this.validRow != null && !this.validRow[d])
                continue;
            for (int i = 0; i < this.numVars; i++) {
                this.priorMean[i] += this.data[d][i];
                for (int j = 0; j < this.numVars; j++)
                    this.priorCov[i][j] += this.data[d][i] * this.data[d][j];
            }
            numValid++;
        }
        for (int i = 0; i < this.numVars; i++) {
            this.priorMean[i] /= numValid;
            for (int j = 0; j < this.numVars; j++)
                this.priorCov[i][j] /= numValid;
        }
        for (int i = 0; i < this.numVars; i++)
            for (int j = 0; j < this.numVars; j++)
                this.priorCov[i][j] -= this.priorMean[i] * this.priorMean[j];
        this.priorInvCov = inverse(this.priorCov);
        this.alpha = 20.;
        this.upsilon = 1;
    }

    public double[][] query(double input[][]) {
        Triple scores[] = new Triple[this.data.length];

        double inputMean[] = new double[this.numVars], S[][] = new double[this.numVars][this.numVars];
        for (int i = 0; i < this.numVars; i++) {
            inputMean[i] = 0.;
            for (int j = 0; j < this.numVars; j++)
                S[i][j] = 0.;
        }
        for (int i = 0; i < input.length; i++)
            for (int v = 0; v < this.numVars; v++)
                inputMean[v] += input[i][v];
        for (int j = 0; j < this.numVars; j++)
            inputMean[j] /= input.length;
        for (int d = 0; d < input.length; d++)
            for (int i = 0; i < this.numVars; i++)
                for (int j = 0; j < this.numVars; j++)
                    S[i][j] += (input[d][i] - inputMean[i]) * (input[d][j] - inputMean[j]);

        this.posteriorAlpha = this.alpha + input.length;
        this.posteriorUpsilon = this.upsilon + input.length;
        for (int j = 0; j < this.numVars; j++)
            this.posteriorMean[j] = (this.upsilon * this.priorMean[j] + input.length * inputMean[j]) /
                                    (this.upsilon + input.length);
        for (int i = 0; i < this.numVars; i++)
            for (int j = 0; j < this.numVars; j++)
                this.posteriorInvCov[i][j] = this.priorInvCov[i][j] + S[i][j] +
                                               (this.upsilon * input.length / (this.upsilon + input.length)) *
                                               (this.priorMean[i] - inputMean[i]) * (this.priorMean[j] - inputMean[j]);


        for (int i = 0; i < this.data.length; i++) {
            scores[i] = new Triple();
            if (this.validRow != null && !this.validRow[i]) {
                scores[i].fields[0] = -Double.MAX_VALUE;
                continue;
            }
            scores[i].fields[0] =  logMarginalLikelihood(this.data[i], this.posteriorMean, this.posteriorUpsilon,
                                                        this.posteriorInvCov, this.posteriorAlpha);
            scores[i].fields[0] -= logMarginalLikelihood(this.data[i], this.priorMean, this.upsilon,
                                                         this.priorInvCov, this.alpha);
            scores[i].fields[1] = i;
            //System.out.println("score (" + (i + 1) + ") = " + scores[i].fields[0] + " [" + i +  "] (out of " +
            //                       this.data.length + " items)");
        }
        Arrays.sort(scores, new TripleComparator());
        double topDocs[][] = new double[this.data.length][2];
        int trim = this.data.length;
        for (int i = 0; i < this.data.length; i++) {
            topDocs[i][0] = (int) scores[i].fields[1];
            topDocs[i][1] = scores[i].fields[0];
            if (scores[i].fields[0] == -Double.MAX_VALUE && trim == this.data.length) {
                trim = i;
                System.out.println("trim = " + trim);
            }
        }
        double trimmedTopDocs[][] = new double[trim][2];
        for (int i = 0; i < trim; i++)
             trimmedTopDocs[i] = topDocs[i];
        return trimmedTopDocs;
    }

    public double[] queryScore(double input[][]) {
        Triple scores[] = new Triple[this.data.length];

        double inputMean[] = new double[this.numVars], S[][] = new double[this.numVars][this.numVars];
        for (int i = 0; i < this.numVars; i++) {
            inputMean[i] = 0.;
            for (int j = 0; j < this.numVars; j++)
                S[i][j] = 0.;
        }
        for (int i = 0; i < input.length; i++)
            for (int v = 0; v < this.numVars; v++)
                inputMean[v] += input[i][v];
        for (int j = 0; j < this.numVars; j++)
            inputMean[j] /= input.length;
        for (int d = 0; d < input.length; d++)
            for (int i = 0; i < this.numVars; i++)
                for (int j = 0; j < this.numVars; j++)
                    S[i][j] += (input[d][i] - inputMean[i]) * (input[d][j] - inputMean[j]);

        this.posteriorAlpha = this.alpha + input.length;
        this.posteriorUpsilon = this.upsilon + input.length;
        for (int j = 0; j < this.numVars; j++)
            this.posteriorMean[j] = (this.upsilon * this.priorMean[j] + input.length * inputMean[j]) /
                                    (this.upsilon + input.length);
        for (int i = 0; i < this.numVars; i++)
            for (int j = 0; j < this.numVars; j++)
                this.posteriorInvCov[i][j] = this.priorInvCov[i][j] + S[i][j] +
                                               (this.upsilon * input.length / (this.upsilon + input.length)) *
                                                  (this.priorMean[i] - inputMean[i]) * (this.priorMean[j] - inputMean[j]);


        for (int i = 0; i < this.data.length; i++) {
            scores[i] = new Triple();
            scores[i].fields[0] =  logMarginalLikelihood(this.data[i], this.posteriorMean, this.posteriorUpsilon,
                                                        this.posteriorInvCov, this.posteriorAlpha);
            scores[i].fields[0] -= logMarginalLikelihood(this.data[i], this.priorMean, this.upsilon,
                                                         this.priorInvCov, this.alpha);
            scores[i].fields[1] = i;
        }
        double scoresOnly[] = new double[this.data.length];
        for (int i = 0; i < this.data.length; i++)
            scoresOnly[i] = scores[i].fields[0];
        return scoresOnly;
    }

    protected static double[][] inverse(double matrix[][]) {
        Algebra algebra = new Algebra();
        DoubleMatrix2D m = new DenseDoubleMatrix2D(matrix);
        DoubleMatrix2D result = algebra.inverse(m);
        return result.toArray();
    }

    public static double determinant(double[][] m) {
        return new Algebra().det(new DenseDoubleMatrix2D(m));
    }

    private static final double COF1 = 76.18009173, COF2 = -86.50532033,  COF3 = 24.01409822,
                                COF4 = -1.231739516, COF5 = 0.120858003e-2, COF6 = -0.536382e-5;

    public static double lngamma(double xx) {



        double x, tmp, ser;



        if (xx < 1.0) {

            return (lngamma(1.0 + xx) - Math.log(xx));

        }

        else {

            x = xx - 1.0;

            tmp = x + 5.5;

            tmp -= (x + 0.5) * Math.log(tmp);

            ser = 1.0 + COF1 / (x + 1.0) + COF2 / (x + 2.0)

                    + COF3 / (x + 3.0) + COF4 / (x + 4.0) + COF5 / (x + 5.0)

                    + COF6 / (x + 6.0);



            return (-tmp + Math.log(2.50662827465 * ser));

        }

    }

    public double logMarginalLikelihood(double point[], double mean[], double upsilon, double invCov[][], double alpha) {
        double logLike = -0.5 * this.numVars * Math.log(2 * Math.PI) - 0.5 * this.numVars * (upsilon / (upsilon + 1))
                         -0.5 * this.numVars * Math.log(2.) + lngamma(0.5 * alpha + 1) - lngamma(0.5 * (alpha + 1 - this.numVars));

        double newInvCov[][] = new double[this.numVars][this.numVars];
        for (int i = 0; i < this.numVars; i++)
            for (int j = 0; j < this.numVars; j++)
                newInvCov[i][j] = invCov[i][j] + (upsilon / (upsilon + 1)) * (mean[i] - point[i]) * (mean[j] - point[j]);

        logLike += 0.5 * alpha * Math.log(determinant(invCov)) - 0.5 * (alpha + 1) * Math.log(determinant(newInvCov));

        return logLike;
    }

    public double[] getMean() {
        return this.priorMean;
    }

    public double[][] getInvCov() {
        return this.priorInvCov;
    }

    public double[][] getCov() {
        return this.priorCov;
    }

    public void setHyperparameters(double mean[], double invCov[][], double alpha, double upsilon) {
        this.priorMean = mean;
        this.priorInvCov = invCov;
        this.alpha = alpha;
        this.upsilon = upsilon;
    }
}
